from PySide6.QtWidgets import QMainWindow, QWidget, QLabel, QLayout
from PySide6.QtGui import QImage, QResizeEvent, QPixmap, QMouseEvent
from PySide6.QtCore import QSize

from abc import abstractmethod
from typing import Optional, List

from .image_entry import ImageEntry
from .checkable_image_entry import CheckableImageEntry
from .utils import load_image, fit_to_frame


class PreviewWindow(QMainWindow):
    def __init__(self, parent: Optional[QWidget]):
        super(PreviewWindow, self).__init__(parent)
        self._selected_image_entry: Optional[ImageEntry] = None
        self._source_image_entries: List[CheckableImageEntry] = []

    @abstractmethod
    def image_preview(self) -> QLabel:
        raise NotImplementedError

    @abstractmethod
    def source_layout(self) -> QLayout:
        raise NotImplementedError

    def add_source_image_entry(self, ime: CheckableImageEntry, index: Optional[int] = None) -> None:
        self._source_image_entries.insert(len(self._source_image_entries) if index is None else index, ime)
        self.source_layout().addWidget(ime)

    def draw_preview_image(self, image: QImage) -> None:
        image_preview = self.image_preview()
        image_preview.setPixmap(
            fit_to_frame(QPixmap.fromImage(image), QSize(image_preview.width(), image_preview.height())))

    def set_preview_image(self, image: QImage, image_entry: ImageEntry) -> None:
        if image_entry is self._selected_image_entry:
            return

        self.draw_preview_image(image)
        if self._selected_image_entry is not None:
            self._selected_image_entry.setSelected(False)
        self._selected_image_entry = image_entry
        image_entry.setSelected(True)

    def clear_preview_image(self) -> None:
        self.image_preview().setText("Preview")
        if self._selected_image_entry is not None:
            self._selected_image_entry.setSelected(False)
        self._selected_image_entry = None

    def image_entry_click_handler(self, sender: ImageEntry, event: QMouseEvent) -> None:
        self.set_preview_image(load_image(sender.image_path), sender)

    def resizeEvent(self, event: QResizeEvent) -> None:
        if self._selected_image_entry is None:
            return
        self.draw_preview_image(load_image(self._selected_image_entry.image_path))

    def clear_image_entries(self) -> None:
        for ime in self._source_image_entries:
            ime.close()
        self._source_image_entries.clear()

    def select_deselect(self) -> None:
        all_checked = True
        for ime in self._source_image_entries:
            if not ime.isChecked():
                all_checked = False
                break

        for ime in self._source_image_entries:
            ime.setChecked(not all_checked)

    def get_selected_entries(self):
        selected: List[CheckableImageEntry] = []
        for ime in self._source_image_entries:
            if ime.isChecked():
                selected.append(ime)
        return selected



from __future__ import annotations

from PySide6.QtWidgets import QWidget
from PySide6.QtGui import QImage, QMouseEvent

from .image_entry import ImageEntry
from .layer_data import LayerData

from typing import Callable, Any, Final, List


class ClusterImageEntry(ImageEntry):
    def __init__(self, parent: QWidget, image: QImage, image_path: str, array_path: str, name: str,
                 layers_data: List[LayerData]):
        """
        Instantiate a ClusterImageEntry object
        :param parent: The widget calling the method
        :param image: The image that will be used to draw the preview thumbnail
        :param image_path: The path of the cluster image
        :param name: The name that will be showed below the thumbnail
        :param layers_data: A list of LayerData containing all the information about a layer
        """
        super(ClusterImageEntry, self).__init__(parent, image, image_path, name, array_path)
        self.layers_data: Final[List[LayerData]] = layers_data
        self.__mouse_double_click_actions: List[Callable[[ClusterImageEntry], Any]] = []

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        for action in self.__mouse_double_click_actions:
            action(self)

    def layer_count(self) -> int:
        return len(self.layers_data)

    def add_layer_data(self, layer_data: LayerData) -> None:
        self.layers_data.append(layer_data)

    def get_layer_data(self, index: int) -> LayerData:
        return self.layers_data[index]

    def remove_layer_data(self, index: int) -> LayerData:
        return self.layers_data.pop(index)

    def register_mouse_double_click_action(self, action: Callable[[ClusterImageEntry], Any]):
        self.__mouse_double_click_actions.append(action)



import os
from pathlib import Path
from typing import Optional, List, Tuple

import numpy as np
import cv2 as cv

from PySide6.QtWidgets import QFileDialog, QLabel, QLayout, QDialog
from PySide6.QtCore import QObject, Slot, Signal, QThreadPool, QRunnable

from .clustering import Tat
from .checkable_image_entry import CheckableImageEntry
from .preview_window import PreviewWindow
from .cluster_image_entry import ClusterImageEntry
from .cluster_editor import ClusterEditor
from .layer_data import LayerData
from .ui_main_window import Ui_MainWindow
from .ui_progress_bar import Ui_ProgressBar
from .utils import load_image, apply_colormap, create_cluster, array3d_to_pixmap


class ClusteringSignals(QObject):
    finished = Signal()
    progress = Signal(int, tuple)


class ClusteringWorker(QRunnable):
    signals = ClusteringSignals()
    is_running = False

    def __init__(self, entries: List[CheckableImageEntry], cluster_count: int, run_count: int, max_iter_count: int,
                 output_directory: str):
        super(ClusteringWorker, self).__init__()
        self.entries = entries
        self.cluster_count = cluster_count
        self.run_count = run_count
        self.max_iter_count = max_iter_count
        self.output_directory = output_directory

    @Slot()
    def run(self) -> None:
        self.is_running = True
        for index, ime in enumerate(self.entries):
            if not self.is_running:
                break

            input_basename_no_ext = (lambda basename: basename[0:basename.rfind(".")])(os.path.basename(ime.image_path))
            layers, cluster = Tat.generate_layers(np.asarray(cv.imread(ime.image_path, flags=cv.IMREAD_GRAYSCALE)),
                                                  self.cluster_count, self.run_count, self.max_iter_count)

            layers_data: List[LayerData] = []
            for i, layer in enumerate(layers):
                output_path_no_ext = os.path.join(self.output_directory, f"{input_basename_no_ext}_layer_{i}")
                output_image_path = f"{output_path_no_ext}.png"
                output_array_path = f"{output_path_no_ext}.npy"
                np.save(output_array_path, layer)
                cv.imwrite(output_image_path, apply_colormap(layer, cv.COLORMAP_VIRIDIS))
                layers_data.append(LayerData(output_image_path, output_array_path, layer_index=i))

            output_path_no_ext = os.path.join(self.output_directory, f"{input_basename_no_ext}_cluster")
            output_image_path = f"{output_path_no_ext}.png"
            output_array_path = f"{output_path_no_ext}.npy"

            np.save(output_array_path, cluster)
            cv.imwrite(output_image_path, apply_colormap(cluster, cv.COLORMAP_JET))

            self.signals.progress.emit(index + 1,
                                       (output_image_path, output_array_path, input_basename_no_ext, layers_data))
        self.signals.finished.emit()

    @Slot()
    def interrupt(self) -> None:
        self.is_running = False


class MainWindow(PreviewWindow):
    def __init__(self):
        super(MainWindow, self).__init__(None)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.thread_pool = QThreadPool()

        self.input_directory = ""
        self.output_directory = ""
        self.merger_directory: Optional[str] = None
        self.editor_window: Optional[ClusterEditor] = None

        self.__generated_images_entries: List[ClusterImageEntry] = []

        self.ui.buttonInputDir.clicked.connect(self.load_input_directory)
        self.ui.buttonOutputDir.clicked.connect(self.load_output_directory)
        self.ui.buttonGenerate.clicked.connect(self.generate_handler)
        self.ui.buttonCheckUncheck.clicked.connect(self.select_deselect)
        self.ui.buttonClearGenerated.clicked.connect(self.clear_generated)

    def source_layout(self) -> QLayout:
        return self.ui.scrollAreaWidgetContentsSrc.layout()

    def image_preview(self) -> QLabel:
        return self.ui.imagePreview

    def clear_image_entries(self):
        super(MainWindow, self).clear_image_entries()
        for ime in self.__generated_images_entries:
            ime.close()
        self.__generated_images_entries.clear()

    def open_preview_window(self, calling_image_entry: ClusterImageEntry):
        if self.editor_window is not None and self.editor_window.isVisible():
            self.editor_window.activateWindow()
            return
        self.editor_window = ClusterEditor(self, calling_image_entry)
        self.editor_window.register_merge_handler(self.merge_layers)
        self.editor_window.show()

    def merge_layers(self, layers_indices: List[int]) -> None:
        """
        Merge all the specified layers
        :param layers_indices: A range of the layers to merge
        :return: None
        """
        if len(layers_indices) == 0:
            return

        merged_cluster_ime: List[ClusterImageEntry] = []

        first = True
        while len(self.__generated_images_entries) > 0:
            ime: ClusterImageEntry = self.__generated_images_entries.pop(0)
            merged: Optional[np.ndarray] = None
            parent_layers: List[int] = []
            for i in layers_indices:
                layer_data: LayerData = ime.get_layer_data(i)
                if layer_data.is_merger:
                    assert layer_data.parent_layers is not None
                    parent_layers.extend(layer_data.parent_layers)
                else:
                    assert layer_data.layer_index is not None
                    parent_layers.append(layer_data.layer_index)

                layer = np.load(layer_data.array_path)
                merged = layer if merged is None else merged | layer

            for i in sorted(layers_indices, reverse=True):
                ime.remove_layer_data(i)

            if merged is None:
                break

            colored = apply_colormap(merged)

            merged_path_no_ext = os.path.join(self.merger_directory,
                                              f"{ime.basename}_layers_{LayerData.indices2str(parent_layers)}")
            merged_image_path = f"{merged_path_no_ext}.png"
            merged_array_path = f"{merged_path_no_ext}.npy"

            cv.imwrite(merged_image_path, colored)
            np.save(merged_array_path, merged)
            ime.add_layer_data(
                LayerData(merged_image_path, merged_array_path, is_merger=True, parent_layers=parent_layers))

            new_cluster_array = create_cluster(
                [np.load(ime.get_layer_data(i).array_path) for i in range(ime.layer_count())])

            new_cluster_colored = apply_colormap(new_cluster_array, cv.COLORMAP_JET)
            new_cluster_colored_image = array3d_to_pixmap(new_cluster_colored).toImage()

            new_cluster_path_no_ext = os.path.join(self.merger_directory, f"{ime.basename}_cluster")
            new_cluster_image_path = f"{new_cluster_path_no_ext}.png"
            new_cluster_array_path = f"{new_cluster_path_no_ext}.npy"

            np.save(new_cluster_array_path, new_cluster_array)
            cv.imwrite(new_cluster_image_path, new_cluster_colored)

            new_ime = ClusterImageEntry(ime.parent(), new_cluster_colored_image, new_cluster_image_path,
                                        new_cluster_array_path, ime.basename, ime.layers_data)
            new_ime.registerMousePressHandler(self.image_entry_click_handler)
            new_ime.register_mouse_double_click_action(self.open_preview_window)
            merged_cluster_ime.append(new_ime)
            ime.close()
            self.ui.scrollAreaWidgetContentsDst.layout().addWidget(new_ime)

            if first:
                self.set_preview_image(new_cluster_colored_image, new_ime)
                first = False

        self.__generated_images_entries = merged_cluster_ime

    def unmerge_layer(self):
        pass

    @Slot()
    def clear_generated(self):
        for ime in self.__generated_images_entries:
            ime.close()
        self.__generated_images_entries.clear()
        self.image_preview().setText("Preview")
        self._selected_image_entry = None

    @Slot()
    def load_input_directory(self):
        in_dir = QFileDialog.getExistingDirectory(self)
        if len(in_dir) == 0:
            return

        self.input_directory = in_dir
        self.clear_image_entries()
        self.clear_preview_image()

        # self.ui.labelInDir.setText(f"Loaded: {self.input_directory}")

        src_layout = self.source_layout()
        first = True
        for entry in sorted(os.scandir(self.input_directory), key=lambda x: x.name):
            if not Tat.is_image(entry.path):
                continue

            qim = load_image(entry.path)

            ime = CheckableImageEntry(src_layout.parent(), qim, entry.name, entry.path)
            ime.registerMousePressHandler(self.image_entry_click_handler)
            self.add_source_image_entry(ime)

            if first:
                self.set_preview_image(qim, ime)
                first = False

        if len(self.output_directory) != 0:
            self.ui.buttonGenerate.setEnabled(True)

    @Slot()
    def load_output_directory(self):
        out_dir = QFileDialog.getExistingDirectory(self)
        if len(out_dir) == 0:
            return

        self.output_directory = out_dir
        self.merger_directory = os.path.join(self.output_directory, "merged")
        Path(self.merger_directory).mkdir(exist_ok=True)

        # self.ui.labelOutDir.setText(f"Loaded: {self.output_directory}")
        if len(self.input_directory) != 0:
            self.ui.buttonGenerate.setEnabled(True)

    @Slot()
    def generate_handler(self):
        self.ui.buttonGenerate.setEnabled(False)

        progress_bar = QDialog(self)
        progress_bar.ui = Ui_ProgressBar()
        progress_bar.ui.setupUi(progress_bar)

        @Slot()
        def add_cluster_image(progress: int, data: Tuple[str, str, str, List[LayerData]]):
            image_path, array_path, name, layers_data = data
            container: QLayout = self.ui.scrollAreaWidgetContentsDst.layout()
            ime = ClusterImageEntry(container.parent(), load_image(image_path), image_path, array_path, name,
                                    layers_data)
            ime.registerMousePressHandler(self.image_entry_click_handler)
            ime.register_mouse_double_click_action(self.open_preview_window)
            container.addWidget(ime)
            if len(self.__generated_images_entries) == 0:
                self.set_preview_image(load_image(ime.image_path), ime)
            progress_bar.ui.progressBar.setValue(progress)
            self.__generated_images_entries.append(ime)

        selected_entries = self.get_selected_entries()
        worker = ClusteringWorker(selected_entries, self.ui.clusterCount.value(), self.ui.runCount.value(),
                                  self.ui.maxIterCount.value(), self.output_directory)

        @Slot()
        def finished_generating():
            progress_bar.close()
            worker.signals.progress.disconnect(add_cluster_image)
            worker.signals.finished.disconnect(finished_generating)
            self.ui.buttonGenerate.setEnabled(True)

        worker.signals.progress.connect(add_cluster_image)
        worker.signals.finished.connect(finished_generating)
        progress_bar.ui.cancelButton.clicked.connect(worker.interrupt)

        progress_bar.ui.progressBar.setMaximum(len(selected_entries))
        progress_bar.show()

        self.thread_pool.start(worker)



from src.tat import MainWindow
from PySide6.QtWidgets import QApplication
import sys

if __name__ == '__main__':
    app = QApplication()
    app.main_window = MainWindow()
    app.main_window.show()
    sys.exit(app.exec())



from PySide6.QtGui import QPixmap, QImage
from PySide6.QtCore import QSize

import numpy as np
import cv2 as cv

from typing import Optional, List


def fit_to_frame(image: QPixmap, frame: QSize) -> QPixmap:
    frame_width, frame_height = frame.toTuple()
    dw = abs(image.width() - frame_width)
    dh = abs(image.height() - frame_height)
    return image.scaledToWidth(frame_width) if dw > dh else image.scaledToHeight(frame_height)


def load_image(path: str) -> QImage:
    return QImage(path)


def apply_colormap(image: np.ndarray, colormap=cv.COLORMAP_VIRIDIS) -> np.ndarray:
    image = image - np.min(image)
    return cv.applyColorMap(np.uint8(255 * (image / np.max(image))), colormap)


def array2d_to_pixmap(array: np.ndarray, normalize=False, colormap: int = cv.COLORMAP_VIRIDIS) -> QPixmap:
    assert array.ndim == 2
    if normalize:
        array = apply_colormap(array, colormap)
        height, width, color_bytes = array.shape
        return QPixmap.fromImage(QImage(array.data, width, height, color_bytes * width, QImage.Format_BGR888))
    height, width = array.shape
    return QPixmap.fromImage(QImage(array.data, width, height, width, QImage.Format_Grayscale8))


def array3d_to_pixmap(array: np.ndarray) -> QPixmap:
    assert array.ndim == 3
    height, width, color_bytes = array.shape
    return QPixmap.fromImage(QImage(array.data, width, height, color_bytes * width, QImage.Format_BGR888))


def create_cluster(layers: List[np.ndarray], normalized=False) -> Optional[np.ndarray]:
    cluster: Optional[np.ndarray] = None
    for index, layer in enumerate(layers):
        cluster = layer.copy() if cluster is None else cluster + layer * (index + 1)
    return cluster / np.max(cluster) if normalized else cluster



from typing import Optional, Final, List

import numpy as np

from .checkable_image_entry import CheckableImageEntry
from .layer_data import LayerData

from PySide6.QtWidgets import QWidget
from PySide6.QtGui import QImage


class LayerImageEntry(CheckableImageEntry):
    def __init__(self, parent: QWidget, image: QImage, array: np.ndarray, name: str, is_merger: bool = False,
                 layer_index: Optional[int] = None, parent_layers: Optional[List[int]] = None):
        super(LayerImageEntry, self).__init__(parent, image, name, default_check=False)

        self.array: Final[np.ndarray] = array
        self.layer_data: Final[LayerData] = LayerData(None, None, is_merger, parent_layers, layer_index)



from .image_entry import ImageEntry
from .cluster_image_entry import ClusterImageEntry
from .checkable_image_entry import CheckableImageEntry
from .layer_image_entry import LayerImageEntry
from .preview_window import PreviewWindow
from .cluster_editor import ClusterEditor
from .main_window import MainWindow
from .layer_data import LayerData



import os
import io
import re
import hashlib
from pbr.git import _iter_log_oneline, _iter_changelog, _run_git_command


def sha256_file(file):
    sha256 = hashlib.sha256()
    with open(file, 'rb') as f:
        sha256.update(f.read())
    return sha256


def generate_changelog(*args, **kwargs):
    changelog = _iter_log_oneline(git_dir='.git')
    if changelog:
        changelog = _iter_changelog(changelog)
    if not changelog:
        return

    new_changelog = os.path.join('docs', 'changelog.md')

    new_content = bytearray()
    for _, content in changelog:
        new_content += content.encode('utf-8')

    if os.path.isfile(new_changelog) and hashlib.sha256(new_content).digest() == sha256_file(new_changelog).digest():
        return

    with io.open(new_changelog, "wb") as changelog_file:
        changelog_file.write(new_content)


def generate_authors(*args, **kwargs):
    old_authors = 'AUTHORS.in'
    new_authors = os.path.join('docs', 'authors.md')
    if os.path.isfile(new_authors) and not os.access(new_authors, os.W_OK):
        # If there's already an AUTHORS file and it's not writable, just use it
        return

    ignore_emails = '((jenkins|zuul)@review|infra@lists|jenkins@openstack)'
    git_dir = '.git'
    authors = []

    # don't include jenkins email address in AUTHORS file
    git_log_cmd = ['log', '--format=%aN <%aE>']
    authors += _run_git_command(git_log_cmd, git_dir).split('\n')
    authors = [a for a in authors if not re.search(ignore_emails, a)]

    # get all co-authors from commit messages
    co_authors_out = _run_git_command('log', git_dir)
    co_authors = re.findall('Co-authored-by:.+', co_authors_out,
                            re.MULTILINE)
    co_authors = [signed.split(":", 1)[1].strip()
                  for signed in co_authors if signed]

    authors += co_authors
    authors = sorted(set(authors))

    new_authors_str = bytearray()

    if os.path.exists(old_authors):
        with open(old_authors, "rb") as old_authors_fh:
            new_authors_str += old_authors_fh.read()
    new_authors_str += (('\n'.join(authors) + '\n').encode('utf-8'))

    if new_authors and hashlib.sha256(new_authors_str).digest() == sha256_file(new_authors).digest():
        return

    with open(new_authors, 'wb') as f:
        f.write(new_authors_str)


def on_pre_build(*args, **kwargs):
    generate_changelog()
    generate_authors()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'progressbar.ui'
##
## Created by: Qt User Interface Compiler version 6.0.4
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *


class Ui_ProgressBar(object):
    def setupUi(self, ProgressBar):
        if not ProgressBar.objectName():
            ProgressBar.setObjectName(u"ProgressBar")
        ProgressBar.resize(400, 100)
        self.verticalLayout = QVBoxLayout(ProgressBar)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.label = QLabel(ProgressBar)
        self.label.setObjectName(u"label")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label.sizePolicy().hasHeightForWidth())
        self.label.setSizePolicy(sizePolicy)
        self.label.setAlignment(Qt.AlignCenter)

        self.verticalLayout.addWidget(self.label)

        self.progressBar = QProgressBar(ProgressBar)
        self.progressBar.setObjectName(u"progressBar")
        self.progressBar.setBaseSize(QSize(0, 0))
        self.progressBar.setMaximum(0)
        self.progressBar.setValue(0)

        self.verticalLayout.addWidget(self.progressBar)

        self.cancelButton = QPushButton(ProgressBar)
        self.cancelButton.setObjectName(u"cancelButton")
        sizePolicy1 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.cancelButton.sizePolicy().hasHeightForWidth())
        self.cancelButton.setSizePolicy(sizePolicy1)
        self.cancelButton.setMinimumSize(QSize(20, 0))

        self.verticalLayout.addWidget(self.cancelButton)


        self.retranslateUi(ProgressBar)

        QMetaObject.connectSlotsByName(ProgressBar)
    # setupUi

    def retranslateUi(self, ProgressBar):
        ProgressBar.setWindowTitle(QCoreApplication.translate("ProgressBar", u"Processing", None))
        self.label.setText(QCoreApplication.translate("ProgressBar", u"Please wait...", None))
        self.progressBar.setFormat(QCoreApplication.translate("ProgressBar", u"%v/%m", None))
        self.cancelButton.setText(QCoreApplication.translate("ProgressBar", u"Cancel", None))
    # retranslateUi




from typing import Optional, Final, List


class LayerData:
    def __init__(self, image_path: str, array_path: str, is_merger=False, parent_layers: Optional[List[int]] = None,
                 layer_index: Optional[int] = None):
        self.image_path: Final[str] = image_path
        self.array_path: Final[str] = array_path
        self.is_merger: Final[bool] = is_merger
        self.parent_layers: Final[Optional[List[int]]] = parent_layers
        self.layer_index: Final[Optional[int]] = layer_index

    def get_name(self) -> str:
        if self.is_merger:
            assert self.parent_layers is not None
            return "m " + self.indices2str(self.parent_layers)
        assert self.layer_index is not None
        return str(self.layer_index)

    @staticmethod
    def indices2str(indices: List[int]) -> str:
        indices_str = ""
        for i in indices:
            indices_str += str(i) if i == indices[0] else f"+{str(i)}"
        return indices_str



from PySide6.QtWidgets import QApplication
from .main_window import MainWindow
import sys


def main() -> int:
    app = QApplication()
    app.main_window = MainWindow()
    app.main_window.show()
    return app.exec()


if __name__ == '__main__':
    sys.exit(main())



from __future__ import annotations
from typing import Callable, Final, Any, Optional, List
from .utils import fit_to_frame

from PySide6.QtWidgets import QWidget, QVBoxLayout, QLabel, QSizePolicy
from PySide6.QtGui import QImage, QPixmap, QColor, QMouseEvent
from PySide6.QtCore import Qt, QSize


class ImageEntry(QWidget):
    def __init__(self, parent: QWidget, image: QImage, image_path: Optional[str], name: str,
                 array_path: Optional[str] = None):
        super(ImageEntry, self).__init__(parent)
        self.__selected = False
        self.__ime_layout = QVBoxLayout(self)
        self.__mouse_pressed_handlers: List[Callable[[ImageEntry, QMouseEvent], Any]] = []
        self.setAutoFillBackground(True)
        self.__default_background_color = self.palette().color(self.backgroundRole())
        self.image_path: Final[Optional[str]] = image_path
        self.array_path: Final[Optional[str]] = array_path
        self.basename: Final[str] = name

        image_label = QLabel("Preview", self)
        image_label.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        image_label.setAlignment(Qt.AlignCenter)
        # qpix = QPixmap.fromImage(image)
        # image_label.setPixmap(qpix.scaledToHeight(image_label.height()))
        image_label.setPixmap(fit_to_frame(QPixmap.fromImage(image), QSize(50, 50)))

        name_label = QLabel(name, self)
        name_label.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        name_label.setAlignment(Qt.AlignCenter)

        self.__ime_layout.addWidget(image_label, alignment=Qt.AlignHCenter)
        self.__ime_layout.addWidget(name_label, alignment=Qt.AlignHCenter)

        self.setLayout(self.__ime_layout)

    def layout(self) -> QVBoxLayout:
        return self.__ime_layout

    def setSelected(self, selected: bool) -> None:
        if selected:
            self.__setBackgroundColor(Qt.gray)
            self.__selected = True
            return

        self.__setBackgroundColor(self.__default_background_color)
        self.__selected = False

    def isSelected(self) -> bool:
        return self.__selected

    def __setBackgroundColor(self, color: QColor):
        pal = self.palette()
        pal.setColor(self.backgroundRole(), color)
        self.setPalette(pal)

    def registerMousePressHandler(self, handler: Callable[[ImageEntry, QMouseEvent], Any]):
        self.__mouse_pressed_handlers.append(handler)

    def clearMousePressHandler(self):
        self.__mouse_pressed_handlers.clear()

    def mousePressEvent(self, event: QMouseEvent) -> None:
        for hdl in self.__mouse_pressed_handlers:
            hdl(self, event)



from PySide6.QtWidgets import QWidget, QCheckBox, QSizePolicy
from PySide6.QtGui import QImage
from PySide6.QtCore import Qt

from .image_entry import ImageEntry

from typing import Optional


class CheckableImageEntry(ImageEntry):
    def __init__(self, parent: QWidget, image: QImage, name: str, image_path: Optional[str] = None,
                 array_path: Optional[str] = None, default_check=True):
        super(CheckableImageEntry, self).__init__(parent, image, image_path, name, array_path)
        layout = self.layout()

        self.__check_box = QCheckBox(self)
        self.__check_box.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Minimum)
        self.__check_box.setChecked(default_check)

        layout.addWidget(self.__check_box, alignment=Qt.AlignHCenter)

    def isChecked(self) -> bool:
        return self.__check_box.isChecked()

    def setChecked(self, checked):
        self.__check_box.setChecked(checked)



from typing import Optional, Callable, Any, Union, List, Tuple

import os

import cv2 as cv
import numpy as np

from PySide6.QtWidgets import QLayout, QLabel, QWidget, QGridLayout
from PySide6.QtGui import QImage, QMouseEvent, QCloseEvent, QResizeEvent, QMoveEvent, QPixmap
from PySide6.QtCore import Slot, QSize, QPoint, Qt, QEvent

from .ui_editor_window import Ui_EditorWindow
from .preview_window import PreviewWindow
from .cluster_image_entry import ClusterImageEntry
from .layer_image_entry import LayerImageEntry
from .layer_data import LayerData
from .utils import load_image, array2d_to_pixmap, fit_to_frame, create_cluster


class CLusterPreviewWindow(QWidget):
    def __init__(self, parent: Optional[QWidget] = None, size: QSize = QSize(600, 600), image: Optional[QImage] = None):
        super(CLusterPreviewWindow, self).__init__(parent)
        self.setWindowFlags(Qt.Window | Qt.FramelessWindowHint)
        self.resize(size)

        self.imageLabel = QLabel("Cluster Preview", self)
        self.imageLabel.setAlignment(Qt.AlignCenter)

        layout = QGridLayout(self)
        layout.addWidget(self.imageLabel)

        if image is not None:
            self.imageLabel.setPixmap(fit_to_frame(QPixmap.fromImage(image), QSize(self.width(), self.height())))

    def update_cluster_preview(self, image: Union[np.ndarray, str]):
        if isinstance(image, np.ndarray):
            self.imageLabel.setPixmap(fit_to_frame(array2d_to_pixmap(image, normalize=True, colormap=cv.COLORMAP_JET),
                                                   QSize(self.width(), self.height())))
            return

        if isinstance(image, str):
            self.imageLabel.setPixmap(
                fit_to_frame(QPixmap.fromImage(QImage(image)), QSize(self.width(), self.height())))


class ClusterEditor(PreviewWindow):
    def __init__(self, parent: Optional[QWidget], calling_image_entry: ClusterImageEntry):
        super(ClusterEditor, self).__init__(parent)
        self.ui = Ui_EditorWindow()
        self.ui.setupUi(self)
        self.ui.mergeButton.clicked.connect(self.merge)
        self.ui.applyButton.clicked.connect(self.apply_to_all)
        self.ui.resetButton.clicked.connect(self.reset)
        # self.ui.unmergeButton.clicked.connect(self.unmerge)
        self.ui.undoButton.clicked.connect(self.undo)

        self.__merge_callback: Optional[Callable[[List[int]], Any]] = None
        self.__unmerge_callback: Optional[Callable] = None

        self._source_image_entries: List[LayerImageEntry] = []
        self._selected_image_entry: Optional[LayerImageEntry] = None
        self.__cluster_image_entry: ClusterImageEntry = calling_image_entry
        self.__pending_mergers: List[List[int]] = []
        self.__pending_ime: List[LayerImageEntry] = []
        self.__old_entries: List[List[LayerImageEntry]] = []
        self.__cluster_array: np.ndarray = np.load(self.__cluster_image_entry.array_path)

        side_length = self.height() - self.menuBar().height()
        self.__cluster_preview_window = CLusterPreviewWindow(self, QSize(side_length, side_length),
                                                             load_image(self.__cluster_image_entry.image_path))
        # self.cluster_preview_window.show()

        # first = True
        for i in range(self.__cluster_image_entry.layer_count()):
            layer_data = self.__cluster_image_entry.get_layer_data(i)
            array = np.load(layer_data.array_path)
            qim: QImage = load_image(layer_data.image_path)
            ime = LayerImageEntry(self, qim, array, layer_data.get_name(), is_merger=layer_data.is_merger,
                                  layer_index=layer_data.layer_index, parent_layers=layer_data.parent_layers)
            ime.registerMousePressHandler(self.image_entry_click_handler)
            self.add_source_image_entry(ime)

            # if first:
            # self.set_preview_image(qim, ime)
            # first = False

    def source_layout(self) -> QLayout:
        return self.ui.scrollAreaLayersContents.layout()

    def image_preview(self) -> QLabel:
        return self.ui.imageLabel

    def register_merge_handler(self, hdl: Callable[[List[int]], Any]):
        self.__merge_callback = hdl

    def image_entry_click_handler(self, sender: LayerImageEntry, event: QMouseEvent) -> None:
        assert type(sender) == LayerImageEntry
        self.set_preview_image(array2d_to_pixmap(sender.array, normalize=True).toImage(), sender)

    def resizeEvent(self, event: QResizeEvent) -> None:
        if self._selected_image_entry is None:
            return
        self.draw_preview_image(array2d_to_pixmap(self._selected_image_entry.array, normalize=True).toImage())

    def moveEvent(self, event: QMoveEvent) -> None:
        position = event.pos()
        self.__cluster_preview_window.move(position - QPoint(self.__cluster_preview_window.width(), 0))
        if self.__cluster_preview_window.isHidden():
            self.__cluster_preview_window.show()

    def closeEvent(self, event: QCloseEvent) -> None:
        self.__cluster_preview_window.close()
        self.__cluster_preview_window = None

    def __pending_add(self, mergers_idx: List[int], ime: LayerImageEntry, old_entries: List[LayerImageEntry]) -> None:
        if not self.ui.undoButton.isEnabled():
            self.ui.undoButton.setEnabled(True)
        self.__pending_mergers.append(mergers_idx)
        self.__pending_ime.append(ime)
        self.__old_entries.append(old_entries)

    def __pending_clear(self) -> None:
        self.ui.undoButton.setEnabled(False)
        self.__pending_mergers.clear()
        self.__pending_ime.clear()
        self.__old_entries.clear()

    def __pending_count(self) -> int:
        return len(self.__pending_mergers)

    def __pending_pop(self) -> Tuple[List[int], LayerImageEntry, List[LayerImageEntry]]:
        if self.__pending_count() == 1:
            self.ui.undoButton.setEnabled(False)
        return self.__pending_mergers.pop(), self.__pending_ime.pop(), self.__old_entries.pop()

    @Slot()
    def merge(self):
        checked_entries: List[int] = []
        old_ime: List[LayerImageEntry] = []
        merger: Optional[np.ndarray] = None
        parent_layers: List[int] = []
        for index, ime in enumerate(self._source_image_entries):
            if not ime.isChecked():
                continue

            if ime.layer_data.is_merger:
                assert ime.layer_data.parent_layers is not None
                parent_layers.extend(ime.layer_data.parent_layers)
            else:
                assert ime.layer_data.layer_index is not None
                parent_layers.append(ime.layer_data.layer_index)

            checked_entries.append(index)
            old_ime.append(ime)
            merger = self._source_image_entries[index].array if merger is None else merger | self._source_image_entries[
                index].array
            ime.setChecked(False)
            ime.close()

        if len(checked_entries) < 2:
            return

        for i in sorted(checked_entries, reverse=True):
            self._source_image_entries.pop(i)

        self.__cluster_array = create_cluster([ime.array for ime in self._source_image_entries])
        self.__cluster_preview_window.update_cluster_preview(self.__cluster_array)

        qim: QImage = array2d_to_pixmap(merger, normalize=True).toImage()
        merged_ime = LayerImageEntry(self, qim, merger, f"m {LayerData.indices2str(parent_layers)}",
                                     is_merger=True, parent_layers=parent_layers)
        merged_ime.registerMousePressHandler(self.image_entry_click_handler)
        self.__pending_add(checked_entries, merged_ime, old_ime)
        self.set_preview_image(qim, merged_ime)
        self.add_source_image_entry(merged_ime)

    @Slot()
    def apply_to_all(self):
        assert self.__merge_callback is not None, "The callback is not defined"
        for merger in self.__pending_mergers:
            self.__merge_callback(merger)
        self.__pending_clear()

    @Slot()
    def reset(self):
        if len(self.__pending_mergers) == 0:
            return

        self.__pending_clear()

        for ime in self._source_image_entries:
            ime.close()

        self._source_image_entries.clear()

        self.__cluster_array = np.load(self.__cluster_image_entry.array_path)
        self.__cluster_preview_window.update_cluster_preview(self.__cluster_array)

        for i in range(self.__cluster_image_entry.layer_count()):
            layer_data = self.__cluster_image_entry.get_layer_data(i)
            array = np.load(layer_data.array_path)
            qim: QImage = load_image(layer_data.image_path)
            ime = LayerImageEntry(self, qim, array, layer_data.get_name(), layer_data.is_merger, layer_data.layer_index,
                                  layer_data.parent_layers)
            ime.registerMousePressHandler(self.image_entry_click_handler)
            self.add_source_image_entry(ime)
            if i == 0:
                self.set_preview_image(load_image(layer_data.image_path), ime)

    @Slot()
    def unmerge(self):
        for index, ime in enumerate(self._source_image_entries):
            if not ime.isChecked() or not ime.layer_data.is_merger:
                continue

            self._source_image_entries.pop(index)
            assert ime.layer_data.parent_layers is not None
            for parent_layer_index in ime.layer_data.parent_layers.copy():
                directory = os.path.dirname(self.__cluster_image_entry.image_path)
                path_no_ext = os.path.join(directory,
                                           f"{self.__cluster_image_entry.basename}_layer_{parent_layer_index}")
                image_path = f"{path_no_ext}.png"
                array_path = f"{path_no_ext}.npy"
                parent_ime = LayerImageEntry(self, load_image(image_path), np.load(array_path), str(parent_layer_index),
                                             layer_index=parent_layer_index)
                parent_ime.registerMousePressHandler(self.image_entry_click_handler)
                self.add_source_image_entry(parent_ime)
            ime.close()

    @Slot()
    def undo(self):
        if self.__pending_count() == 0:
            return

        indices, pending_ime, old_ime = self.__pending_pop()
        self._source_image_entries.pop()

        pending_ime.close()

        for index, ime in zip(indices, old_ime):
            ime.setVisible(True)
            self.add_source_image_entry(ime, index)
        self.image_preview().setText("Layer")
        self.__cluster_preview_window.update_cluster_preview(self.__cluster_image_entry.image_path)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'clustereditor.ui'
##
## Created by: Qt User Interface Compiler version 6.0.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import *
from PySide6.QtGui import *
from PySide6.QtWidgets import *


class Ui_EditorWindow(object):
    def setupUi(self, EditorWindow):
        if not EditorWindow.objectName():
            EditorWindow.setObjectName(u"EditorWindow")
        EditorWindow.resize(800, 600)
        self.centralwidget = QWidget(EditorWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.gridLayout_2 = QGridLayout(self.centralwidget)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.undoButton = QPushButton(self.centralwidget)
        self.undoButton.setObjectName(u"undoButton")
        self.undoButton.setEnabled(False)

        self.gridLayout_2.addWidget(self.undoButton, 2, 1, 1, 1)

        self.mergeButton = QPushButton(self.centralwidget)
        self.mergeButton.setObjectName(u"mergeButton")

        self.gridLayout_2.addWidget(self.mergeButton, 1, 1, 1, 1)

        self.applyButton = QPushButton(self.centralwidget)
        self.applyButton.setObjectName(u"applyButton")

        self.gridLayout_2.addWidget(self.applyButton, 8, 1, 1, 1)

        self.imageFrame = QFrame(self.centralwidget)
        self.imageFrame.setObjectName(u"imageFrame")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.imageFrame.sizePolicy().hasHeightForWidth())
        self.imageFrame.setSizePolicy(sizePolicy)
        self.imageFrame.setFrameShape(QFrame.StyledPanel)
        self.imageFrame.setFrameShadow(QFrame.Raised)
        self.gridLayout = QGridLayout(self.imageFrame)
        self.gridLayout.setObjectName(u"gridLayout")
        self.imageLabel = QLabel(self.imageFrame)
        self.imageLabel.setObjectName(u"imageLabel")
        self.imageLabel.setMinimumSize(QSize(300, 300))
        self.imageLabel.setAlignment(Qt.AlignCenter)

        self.gridLayout.addWidget(self.imageLabel, 0, 0, 1, 1)


        self.gridLayout_2.addWidget(self.imageFrame, 0, 0, 9, 1)

        self.resetButton = QPushButton(self.centralwidget)
        self.resetButton.setObjectName(u"resetButton")

        self.gridLayout_2.addWidget(self.resetButton, 0, 1, 1, 1)

        self.scrollAreaLayers = QScrollArea(self.centralwidget)
        self.scrollAreaLayers.setObjectName(u"scrollAreaLayers")
        sizePolicy1 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Expanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.scrollAreaLayers.sizePolicy().hasHeightForWidth())
        self.scrollAreaLayers.setSizePolicy(sizePolicy1)
        self.scrollAreaLayers.setMinimumSize(QSize(150, 0))
        self.scrollAreaLayers.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.scrollAreaLayers.setWidgetResizable(True)
        self.scrollAreaLayersContents = QWidget()
        self.scrollAreaLayersContents.setObjectName(u"scrollAreaLayersContents")
        self.scrollAreaLayersContents.setGeometry(QRect(0, 0, 148, 452))
        self.verticalLayout = QVBoxLayout(self.scrollAreaLayersContents)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.scrollAreaLayers.setWidget(self.scrollAreaLayersContents)

        self.gridLayout_2.addWidget(self.scrollAreaLayers, 3, 1, 5, 1)

        EditorWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(EditorWindow)

        QMetaObject.connectSlotsByName(EditorWindow)
    # setupUi

    def retranslateUi(self, EditorWindow):
        EditorWindow.setWindowTitle(QCoreApplication.translate("EditorWindow", u"Cluster Editor", None))
        self.undoButton.setText(QCoreApplication.translate("EditorWindow", u"Undo", None))
        self.mergeButton.setText(QCoreApplication.translate("EditorWindow", u"Merge", None))
        self.applyButton.setText(QCoreApplication.translate("EditorWindow", u"Apply to all (save)", None))
        self.imageLabel.setText(QCoreApplication.translate("EditorWindow", u"Layer", None))
        self.resetButton.setText(QCoreApplication.translate("EditorWindow", u"Reset", None))
    # retranslateUi

